//
//    Title: painter.cc
//  Created: Tue Oct  2 15:53:57 2001
//   Author: Tim Weyrich <weyrich@inf.ethz.ch>
//
//  copyright (c) 2001--2003, Computer Graphics Laboratory, ETH Zuerich
//
//
// ========================================================================
// #     #
// ##   ##  #    #   ####  #####  ######  #####   #       ####    ####
// # # # #  #    #  #        #    #       #    #  #      #       #    #
// #  #  #  #    #   ####    #    #####   #    #  #       ####   #
// #     #  #    #       #   #    #       #####   #           #  #  ###
// #     #  #    #  #    #   #    #       #   #   #      #    #  #    #  ##
// #     #   ####    ####    #    ######  #    #  ######  ####    ####   ##
// ========================================================================
//

#include <stdlib.h>
#include <iostream.h>
#include <math.h>

#ifndef M_PI
#  define  M_PI  3.14159265358979323846     // in case pi wasn't defined in math.h
#endif

// Header for the application framework:
#include <framewk.h>

//
//  See http://graphics.ethz.ch/~weyrich/Info-I/painter/ for a
//  program documentation...
//

ImageGray  *myImage = NULL;

ImageGray  *myImageBuffer = NULL;

bool   mousePressed = false;
float  mouseX = -1.0f;
float  mouseY = -1.0f;

// Line drawing.
static int lineX0, lineY0, lineX1, lineY1;

enum
  {
    PNT_NONE,
    // Original features from the framework:
    PNT_DOTS, PNT_HOOVER,
    // Exercise 1:
    PNT_LINE,
    // Exercise 2:
    PNT_KALEIDOSCOPE_SQUARE,
    // A more sophisticated variant (not part of the exercise):
    PNT_KALEIDOSCOPE_POLYGON,
    // Exercise 3:
    PNT_INTERACTIVE_MIRROR
  }
paintMode = PNT_LINE;

///////////////////////////////////////////////////////////////////////////////
// A Sample Effect Called `Hoover'                                           //
///////////////////////////////////////////////////////////////////////////////

void  hoover(ImageGray *img, float mouseX, float mouseY)
  //! Applies my `hoover effect' to an image.
  /*! This function implements a nice effect on an image. It is not
   *  necessary, that you understand this code in total. But for the
   *  interested, it may show, that even more spectacular actions on
   *  images may be coded easily...
   */
{
  // Hints to understand the code:
  //
  //   rand() returns an integer out of 0..RAND_MAX. RAND_MAX is a
  //   very large integer.
  //
  // 
  //   Using rand(),
  const int   w = img->width;
  const int   h = img->height;

  if (img!=NULL && mousePressed)
    {
      for (int i=50000; i; i--)
        {
          double  phi = 2.0*3.14159*(rand() & 32767)/32768;
          double  r = rand() % w;
          
          double  c = cos(phi);
          double  s = sin(phi);

          int  x1 = (int)(mouseX + r * c);
          int  y1 = (int)(mouseY + r * s);

          if (x1 >= 0 && x1 < w && y1 >= 0 && y1 < h)
            {
              int  x2 = (int)(mouseX + (r-1.0) * c);
              int  y2 = (int)(mouseY + (r-1.0) * s);
              
              img->data[x2 + y2*w] = img->data[x1 + y1*w];
            }
        }
    }
}

///////////////////////////////////////////////////////////////////////////////
// Kaleidoscope_square
///////////////////////////////////////////////////////////////////////////////

inline int  multipleMirror(const int x, const int period, const int period2)
{
  int  u = x % period2;
  if (u < 0)  u = u + period2;
  if (u < period)
    return u;
  else
    return period2 - u;
}

void  kaleidoscope_square(ImageGray *src, int x, int y, ImageGray *dest)
{
  int  s, t;
  int  u, v;
  int  srcIdx, destIdx;

  const int  tileWidth = 96;
  const int  tileHeight = 96;

  const int  tileWidth2 = 2*tileWidth;
  const int  tileHeight2 = 2*tileHeight;

  x = x - tileWidth / 2;
  y = y - tileHeight / 2;
  
  for (t=0; t<dest->height; t++)
    for (s=0; s<dest->width; s++)
      {
        u = x + multipleMirror(s-x, tileWidth, tileWidth2);
        v = y + multipleMirror(t-y, tileHeight, tileHeight2);
        //cerr << u << ", " << v << endl;

        if (u < 0)  u = 0;
        if (v < 0)  v = 0;

        if (u >= src->width)   u = src->width-1;
        if (v >= src->height)  v = src->height-1;

        srcIdx = u + v*src->width;
        destIdx = s + t*dest->width;

        dest->data[destIdx] = src->data[srcIdx];
      }
  
  if (false)  // change it to `true' to enable the square border
    {
      unsigned char  black = 0;
      unsigned char  white = 255;
    
      imageGrayDrawLine(dest, x, y, x+tileWidth, y, white);
      imageGrayDrawLine(dest, x, y+tileHeight, x+tileWidth, y+tileHeight, white);
      imageGrayDrawLine(dest, x, y, x, y+tileHeight, white);
      imageGrayDrawLine(dest, x+tileWidth, y, x+tileWidth, y+tileHeight, white);
    
      imageGrayDrawLine(dest, x-1, y-1, x+tileWidth-1, y-1, black);
      imageGrayDrawLine(dest, x-1, y+tileHeight-1, x+tileWidth-1, y+tileHeight-1, black);
      imageGrayDrawLine(dest, x-1, y-1, x-1, y+tileHeight-1, black);
      imageGrayDrawLine(dest, x+tileWidth-1, y, x+tileWidth-1, y+tileHeight-1, black);
    }
}

///////////////////////////////////////////////////////////////////////////////
// Kaleidoscope_polygon
///////////////////////////////////////////////////////////////////////////////

int    kalPolN;
int    kalPolWidth;
int    kalPolHeight;
long  *kalPolOffs = NULL;
short *kalPolX = NULL;
unsigned char  *kalPolMaskIntensity = NULL;

const float  kalPolRotCorr = 0.125;

const bool  pseudo3DBorder = true;

void  initKaleidoscope_polygon(int n, int width, int height)
{
  if (kalPolOffs)
    {
      if (n == kalPolN && width == kalPolWidth && height == kalPolHeight)
        return;

      delete[] kalPolOffs;
      delete[] kalPolX;
      delete[] kalPolMaskIntensity;
    }

  kalPolWidth = width;
  kalPolHeight = height;
  kalPolN = n;
  
  kalPolOffs = new long[4*width*height];
  kalPolX = new short[4*width*height];
  kalPolMaskIntensity = new unsigned char[4*width*height];

  {
    int  w = 2*kalPolWidth;
    int  h = 2*kalPolHeight;
  
    int    s, t;
    
    const float  r = 45;  // radius of the inscribing circle
    
    const double  pi2 = 6.28318530717959;
    
    float  *cos_tab = new float[n+1];
    float  *sin_tab = new float[n+1];
    float  *cos_tab2 = new float[n+1];
    float  *sin_tab2 = new float[n+1];
    
    float  u, v;
    float  phi;
    float  cs, sn;
    float  d, b;
    
    const float  border = 0.3;
    float        borderAttenuation;
    
    int  i, m;
    
    for (i=0; i<n+1; i++)
      {
        cos_tab[i] = cos(pi2 * (i + kalPolRotCorr) / n);
        sin_tab[i] = sin(pi2 * (i + kalPolRotCorr) / n);
        cos_tab2[i] = cos(pi2 * (i + kalPolRotCorr - 0.5) / n);
        sin_tab2[i] = sin(pi2 * (i + kalPolRotCorr - 0.5) / n);
      }
    
    cerr << "precomputing kaleidoscope effect..";
    for (t=0; t<h; t++)
      for (s=0; s<w; s++)
        {
          u = s-kalPolWidth;
          v = t-kalPolHeight;
          borderAttenuation = 1.f;
          for (m=0; ; m++)
            {
              phi = atan2(v, u);
              if (phi < 0)  phi += pi2;
              i = (int)floor((n * phi / pi2) + 0.5 - kalPolRotCorr);
              if (i >= n)  i -= n;

              cs = cos_tab[i];
              sn = sin_tab[i];
          
              d = u*cs + v*sn;
              if (d <= r + 1e-5)   // safety epsilon
                {
                  b = r - d;
                  if (b < border+0.5)
                    if (b > border-0.5)
                      borderAttenuation *= (b - (border-0.5));
                    else
                      borderAttenuation = 0.0;
          
                  break;
                }

              float  b1 = (float)fabs(v*cos_tab2[i] - u*sin_tab2[i]);
              float  b2 = (float)fabs(v*cos_tab2[i+1] - u*sin_tab2[i+1]);
              b = b1 < b2 ? b1 : b2;
              float  brd = border;
              if (pseudo3DBorder)
                brd *= d / r;
              if (b < brd+0.5)
                if (b > brd-0.5)
                  borderAttenuation *= (b - (brd-0.5));
                else
                  borderAttenuation = 0.0;
          
              d = 2*(r - d);
              u = u + d*cs;
              v = v + d*sn;
            }

          kalPolX[s + t*w] = (short)floor(u+0.5);
          kalPolOffs[s + t*w] = (long)(floor(u+0.5) + width*floor(v+0.5));
          kalPolMaskIntensity[s + t*w] = (unsigned char)floor(borderAttenuation * pow(0.75, m) * 255 + 0.5);
        }
    cerr << ".done.\n";
    
    delete[] cos_tab;
    delete[] sin_tab;
    delete[] cos_tab2;
    delete[] sin_tab2;
  }
}

void  kaleidoscope_polygon(ImageGray *src, int x, int y, ImageGray *dest)
{
  int    s, t;
  int    kalPolIdx;
  long   originOffs = x + y*dest->width;
  long   endOffs = dest->width * dest->height;

  int  srcIdx, destIdx;

  kalPolIdx = (dest->width - x) + 2*dest->width*(dest->height - y);
  destIdx = 0;
  for (t=0; t<dest->height; t++, kalPolIdx+=dest->width)
    for (s=0; s<dest->width; s++, kalPolIdx++, destIdx++)
      {
        const int  X = kalPolX[kalPolIdx] + x;
        srcIdx = originOffs + kalPolOffs[kalPolIdx];
        if (X >= 0 && X < dest->width &&
            srcIdx >= 0 &&
            srcIdx < endOffs)
          {
            const unsigned int  f = kalPolMaskIntensity[kalPolIdx];
            dest->data[destIdx] = (unsigned char)(f * src->data[srcIdx] / 255);
          }
        else
          dest->data[destIdx] = 0; // black
      }
}

///////////////////////////////////////////////////////////////////////////////
// Interactive Mirror (Exercise 3)
///////////////////////////////////////////////////////////////////////////////

unsigned char  getPixel_nearest(ImageGray *img, float x, float y)
{
  int  s = (int)floor(x + 0.5);
  int  t = (int)floor(y + 0.5);

  if (s >= 0 && t >= 0 &&
      s < img->width && t < img->height)
    {
      return img->data[s +  t*img->width];
    }
  else
    {
      return 0;  // return black
    }
}

unsigned char  getPixel_bilinear(ImageGray *img, float x, float y)
{
  float  u = floor(x);
  float  v = floor(y);

  int  s = (int)u;
  int  t = (int)v;

  u = x - u;
  v = y - v;

  const unsigned char  p1 = getPixel_nearest(img, s,   t);
  const unsigned char  p2 = getPixel_nearest(img, s+1, t);
  const unsigned char  p3 = getPixel_nearest(img, s,   t+1);
  const unsigned char  p4 = getPixel_nearest(img, s+1, t+1);

  return (unsigned char)((1-u)*(1-v)*p1 + u*(1-v)*p2 + (1-u)*v*p3 + u*v*p4);
}

void  interactiveMirror(ImageGray *src, float x0, float y0, float x1, float y1,
                        ImageGray *dest)
{
#if 1
  // This is the version using the formulae as described in the
  // exercise. The #else-branches presents alternative solutions that
  // can be ignored.
  //
  // The solution proposed in the exercise is presumably the one with
  // the minimum of floating point operations (FLOPS) required.
  //
  {
    float  ab[2] = {x1-x0, y1-y0};
    float  abDist = (float)sqrt(ab[0]*ab[0] + ab[1]*ab[1]);
    
    float  n[2] = {ab[1]/abDist, -ab[0]/abDist};

    int    x, y;  // destination pixel
    float  u, v;  // source position
    
    for (y=0; y<dest->height; y++)
      for (x=0; x<dest->width; x++)
        {
          float  d = (x-x0)*n[0] + (y-y0)*n[1];

          if (d > 0)
            {
              // Now compute uv = xy - 2*d*n:
              d *= 2;
              u = x - d*n[0];
              v = y - d*n[1];
        
              dest->data[x + y*dest->width] = getPixel_bilinear(src, u, v);
            }
          else
            {
              // We assume that the destination image already contains
              // the source image, so here there is nothing to do...
            }
        }
  }
#else
  // Alternative solutions
  //
  {
    int    x, y;
    float  u, v;
  
    float  dx = x1 - x0;
    float  dy = y1 - y0;
  
    double  d = sqrt(dx*dx + dy*dy);
  
    dx /= d;
    dy /= d;

#  if 1
    const float  toOrigin[2][2] = {{-dy, dx}, { dx,  dy}};
    const float  toMouse[2][2] =  {{-dy, dx}, { dx,  dy}};
  
    for (y=0; y<dest->height; y++)
      for (x=0; x<dest->width; x++)
        {
          u = toOrigin[0][0]*(x-x0) + toOrigin[0][1]*(y-y0);
          v = toOrigin[1][0]*(x-x0) + toOrigin[1][1]*(y-y0);
        
          if (u < 0)
            u = -u;
        
          dest->data[x + y*dest->width]
            = getPixel_bilinear(src,
                                toMouse[0][0]*u + toMouse[0][1]*v + x0,
                                toMouse[1][0]*u + toMouse[1][1]*v + y0);
        }
#  else
    // This is a somewhat "generic" solution using a general affine
    // transformation uv = A*xy + t.
    //
    const float  A[2][2] = {{dx*dx-dy*dy, 2*dx*dy}, {2*dx*dy, dy*dy-dx*dx}};
    const float  t[2] = {2*dy*(-dx*y0 + x0*dy), -2*dx*(-dx*y0 + x0*dy)};
  
    for (y=0; y<dest->height; y++)
      for (x=0; x<dest->width; x++)
        {
          u = A[0][0]*x + A[0][1]*y + t[0];
          v = A[1][0]*x + A[1][1]*y + t[1];
        
          dest->data[x + y*dest->width] = getPixel_bilinear(src, u, v);
        }
#  endif
  }
#endif
}

///////////////////////////////////////////////////////////////////////////////
// Image Rotation (Exercise 3)                                               //
///////////////////////////////////////////////////////////////////////////////

void  rotateImage(ImageGray *src, float x0, float y0, float phiDegree,
                  ImageGray *dest, bool bilinearItp)
{
  int    x, y;
  float  u, v;
  
  float  phi = - M_PI * phiDegree / 180;
  
  if (bilinearItp)
    for (y=0; y<dest->height; y++)
      for (x=0; x<dest->width; x++)
        {
          u = x0 + (x - x0) * cos(phi) - (y - y0) * sin(phi);
          v = y0 + (x - x0) * sin(phi) + (y - y0) * cos(phi) ;
          
          dest->data[x + y*dest->width]
            = getPixel_bilinear(src, u, v);
        }
  else
    for (y=0; y<dest->height; y++)
      for (x=0; x<dest->width; x++)
        {
          u = x0 + (x - x0) * cos(phi) - (y - y0) * sin(phi);
          v = y0 + (x - x0) * sin(phi) + (y - y0) * cos(phi) ;
          
          dest->data[x + y*dest->width]
            = getPixel_nearest(src, u, v);
        }
}

///////////////////////////////////////////////////////////////////////////////
// Callback Functions                                                        //
///////////////////////////////////////////////////////////////////////////////

void  keyboardCB(unsigned char key, float x, float y)
  /*! This callback function is called, when a non-special key (see
   *  specialCB()) was pressed.
   *
   *   \param key    The ASCII code of the key, that was pressed.
   *
   *   \param x, y   Mouse position, when the key was pressed.
   *                 The position is measured in image coordinates.
   *                 Note that x and y can be fractional numbers if
   *                 the image is zoomed.  If the mouse is outside the
   *                 window, it may happen that x and y are outside
   *                 the image boundaries.
   */
{
  (void)x; (void)y; // This line prevents the compiler warning `variables not used'.
  
  switch (key)
    {
    case 'o':
    case 'O':
      {
        ImageGray *buffer = imageGrayClone(myImage);
        rotateImage(buffer, 265, 265, 15, myImage, key == 'o');
        imageGrayFree(buffer);
        fwkPostRedisplay();
      }
      break;

    case 'r':
      if (myImage)
        imageGrayFree(myImage);
      myImage = imageGrayReadTIFF("in.tif");
      fwkSetDisplayImage(myImage);
      fwkPostRedisplay();
      break;
      
    case 's':
      imageGrayWriteTIFF("out.tif", myImage);
      break;
      
    case '+':
    case '=':
      fwkSetDisplayZoom(2 * fwkGetDisplayZoom());
      break;
      
    case '-':
      fwkSetDisplayZoom(0.5 * fwkGetDisplayZoom());
      break;
      
    case 'q':
      exit(0);
      break;

    case '?':
      cout
        << "" << endl
        << "  Keyboard Shortcuts:" << endl
        << "  -----------------------------------------------------------------------------" << endl
        << "" << endl
        << "  o     Rotate image, bilinear interpolation (exercise 3)" << endl
        << "  O     Rotate image, nearest neighbour interpolation (exercise 3)" << endl
        << "" << endl
        << "  F1    Line mode (exercise 1)" << endl
        << "  F2    Kaleidoscope (exercise 2)" << endl
        << "  F3    Interactive mirror (exercise 3)" << endl
        << "  F10   Dots Mode from original framework" << endl
        << "  F11   Hoover Mode from original framework" << endl
        << "  F12   Tim's special kaleidoscope..." << endl
        << "" << endl
        << "  s     Save image as `out.tif'." << endl
        << "  +/-   Zoom in/out." << endl
        << "" << endl
        << "  q     Quit this application." << endl
        << "  ?     This help." << endl
        << "" << endl;
      break;
      
    default:
      cout << "\aKey `" << key << "' (code " << (int)key << ") is not defined." << endl;
    }
}

void  specialCB(int key, float x, float y)
  /*! This callback function is called, when a `special key', i.e. a
   *  key with no printable representation, was pressed.
   *
   *   \param key    The special code of the key, that was pressed.
   *                 Symbolic constants for the possible values of
   *                 key can be found in GL/glut.h. (On the Sun
   *                 computers in the exercise pool, this file is
   *                 located at
   *                 /usr/pack/glut-3.7-mo/solaris/include/GL/glut.h.)
   *
   *   \param x, y   Mouse position, when the key was pressed.
   *                 The position is measured in image coordinates.
   *                 Note that x and y can be fractional numbers if
   *                 the image is zoomed.  If the mouse is outside the
   *                 window, it may happen that x and y are outside
   *                 the image boundaries.
   */
{
  (void)x; (void)y; // This line prevents the compiler warning `variables not used'.

  switch(key)
    {
    case GLUT_KEY_F1:           // Exercise 1
      paintMode = PNT_LINE;
      cout << "Line Mode." << endl;
      break;

    case GLUT_KEY_F2:           // Exercise 2
      paintMode = PNT_KALEIDOSCOPE_SQUARE;
      cout << "Kaleidoscope Mode (Square)." << endl;
      break;

    case GLUT_KEY_F3:           // Exercise 3
      paintMode = PNT_INTERACTIVE_MIRROR;
      cout << "Interactive Kaleidoscope Mode." << endl;
      break;

    case GLUT_KEY_F10:          // Dots mode from original framework
      paintMode = PNT_DOTS;
      cout << "Dots Mode." << endl;
      break;

    case GLUT_KEY_F11:          // Hoover effect from original framework
      paintMode = PNT_HOOVER;
      cout << "Hoover Mode." << endl;
      break;

    case GLUT_KEY_F12:          // A more advanced kaleidoscope
      {
        const int  n = 5;
        paintMode = PNT_KALEIDOSCOPE_POLYGON;
        cout << "Kaleidoscope Mode (Polygonal, n=" << n << ")." << endl;
        initKaleidoscope_polygon(n, myImage->width, myImage->height);
      }
      break;

    default:
      cout << "\aKey not defined." << endl;
    }
}

void  mouseCB(int button, int state, float x, float y)
  /*! This callback function is called, when a mouse button was
   *  pressed or released.
   *
   *   \param button         The number of the button, that was
   *                         pressed. \a button can be one of
   *                         the following values: GLUT_LEFT_BUTTON,
   *                         GLUT_MIDDLE_BUTTON, GLUT_RIGHT_BUTTON.
   *
   *   \param state          Denotes, whether the button was
   *                         pressed or released. \a state can be
   *                         be one of the following values:
   *                         GLUT_DOWN, when the button was pressed,
   *                         or GLUT_UP, when the button was
   *                         released.
   *
   *   \param x, y           Mouse position, when the button was
   *                         pressed/released.
   *                         The position is measured in image
   *                         coordinates. Note that x and y can be
   *                         fractional numbers if the image is
   *                         zoomed.  If the mouse is outside the
   *                         window, it may happen that x and y are
   *                         outside the image boundaries.
   */
{
  // TODO: Insert your Code, that reacts on mouse button actions. The
  // following code demonstrates the meaning of `button' and `state':
  //
  if (button == GLUT_LEFT_BUTTON)
    if (state == GLUT_DOWN)
      {
        // Save current image coordinates in mouseX/mouseY.
        mouseX = x;
        mouseY = y;
        mousePressed = true;

        switch (paintMode)
          {
          case PNT_NONE:
            cout << "Mouse @ (" << mouseX << "," << mouseY << ")" << endl;
            break;
                          
          case PNT_DOTS:
            if (mouseX >= 0 && mouseY >= 0 &&
                mouseX < myImage->width && mouseY < myImage->height)
              { // Nicky's dots:
                int  Ix = (int)mouseX;
                int  Iy = (int)mouseY;
                myImage->data[Iy*myImage->width+Ix] = 0;
                fwkPostRedisplay();
              }
            break;

          case PNT_LINE:
          case PNT_INTERACTIVE_MIRROR:
            // Set first vertex for line drawing.
            lineX0 = lineX1 = (int)floor(mouseX + 0.5);
            lineY0 = lineY1 = (int)floor(mouseY + 0.5);
            break;
                          
          case PNT_KALEIDOSCOPE_SQUARE:
            if (mouseX >= 0 && mouseY >= 0 &&
                mouseX < myImage->width && mouseY < myImage->height)
              {
                if (myImageBuffer)  imageGrayFree(myImageBuffer);
                myImageBuffer = imageGrayClone(myImage);
                kaleidoscope_square(myImageBuffer, (int)floor(mouseX+0.5), (int)floor(mouseY+0.5),
                                    myImage);
                fwkPostRedisplay();
              }
            break;
                        
          case PNT_KALEIDOSCOPE_POLYGON:
            if (mouseX >= 0 && mouseY >= 0 &&
                mouseX < myImage->width && mouseY < myImage->height)
              {
                if (myImageBuffer)  imageGrayFree(myImageBuffer);
                myImageBuffer = imageGrayClone(myImage);
                kaleidoscope_polygon(myImageBuffer, (int)floor(mouseX+0.5), (int)floor(mouseY+0.5),
                                     myImage);
                fwkPostRedisplay();
              }
            break;

          default:
            break;
          }
      }
    else
      {
        mouseX = -1.0f;
        mouseY = -1.0f;
        mousePressed = false;

        switch (paintMode)
          {
          case PNT_LINE:
            {
              if ( (lineX0 != lineX1) || (lineY0 != lineY1) )
                {
                  unsigned char  grayValue = 128;
                  imageGrayInvertLine(myImage, lineX0, lineY0, lineX1, lineY1);
                  imageGrayDrawLine(myImage, lineX0, lineY0, lineX1, lineY1, grayValue);
                }
              fwkPostRedisplay();
            }
            break;
                          
          case PNT_INTERACTIVE_MIRROR:
            {
              if ( (lineX0 != lineX1) || (lineY0 != lineY1) )
                {
                  imageGrayInvertLine(myImage, lineX0, lineY0, lineX1, lineY1);

                  ImageGray  *buffer = imageGrayClone(myImage);
                  interactiveMirror(buffer, lineX0, lineY0, lineX1, lineY1, myImage);
                  imageGrayFree(buffer);
                }
              fwkPostRedisplay();
            }
            break;

          default:
            break;
          }
      }
}

void  motionCB(float x, float y)
  /*! This callback function is called, when the mouse is moved,
   *  while a mouse button is held pressed.
   *
   *   \param x, y           Current mouse position.
   *                         The position is measured in image
   *                         coordinates. Note that x and y can be
   *                         fractional numbers if the image is
   *                         zoomed.  If the mouse is outside the
   *                         window, it may happen that x and y are
   *                         outside the image boundaries.
   */
{
  if (mousePressed)
    {
      mouseX = x;
      mouseY = y;
      
      if (paintMode == PNT_DOTS &&
          mouseX >= 0 && mouseY >= 0 &&
          mouseX < myImage->width && mouseY < myImage->height)
        { // Nicky's dots:
          int  Ix = (int)mouseX;
          int  Iy = (int)mouseY;
          myImage->data[Iy*myImage->width+Ix] = 0;
          fwkPostRedisplay();
        }
      else if (paintMode == PNT_LINE || paintMode == PNT_INTERACTIVE_MIRROR)
        {
          // Visualize selected line.
          if ( (lineX0 != lineX1) || (lineY0 != lineY1) )
            {
              imageGrayInvertLine(myImage, lineX0, lineY0, lineX1, lineY1);
            }

          lineX1 = (int)floor(mouseX + 0.5);
          lineY1 = (int)floor(mouseY + 0.5);

          if ( (lineX0 != lineX1) || (lineY0 != lineY1) )
            {
              imageGrayInvertLine(myImage, lineX0, lineY0, lineX1, lineY1);
            }

          fwkPostRedisplay();
        }
    }
}

void  idleCB(void)
  /*! This callback function is called, whenever the application
   *  main-loop is idle.
   */
{
  switch(paintMode)
    {
    case PNT_HOOVER:
      if (mousePressed)
        {
          hoover(myImage, mouseX, mouseY);
          fwkPostRedisplay();
        }
      break;

    case PNT_KALEIDOSCOPE_SQUARE:
      if (mousePressed)
        {
          kaleidoscope_square(myImageBuffer, (int)floor(mouseX+0.5), (int)floor(mouseY+0.5), myImage);
          fwkPostRedisplay();
        }
      break;
      
    case PNT_KALEIDOSCOPE_POLYGON:
      if (mousePressed)
        {
          kaleidoscope_polygon(myImageBuffer, (int)floor(mouseX+0.5), (int)floor(mouseY+0.5), myImage);
          fwkPostRedisplay();
        }
      break;

    default:
      break;
    }
}


///////////////////////////////////////////////////////////////////////////////
// main                                                                      //
///////////////////////////////////////////////////////////////////////////////

int  main(int argc, const char *argv[])
  /*! See \ref pgdocFramework for information, how to write your
   *  main() function.
   */
{
  // Important: Initialize application framework.
  fwkInit(&argc, argv);

  {
    // TODO: Do your own initializing stuff. The following lines are
    // an example:
    myImage = imageGrayReadTIFF("in.tif");
    fwkSetDisplayImage(myImage);
  }

  // Start application main-loop.
  fwkMainLoop();
  
  return 1; // Return an error value, since we should never reach this point!
}

///////////////////////////////////////////////////////////////////////////////
